lsdupes - a lister of duplicate files
=====================================
Copyright 2006-2010 Ben Clifford
BSD-like license

> import Control.Applicative
> import Control.Monad
> import Control.Monad.Writer
> import Data.IORef
> import Data.List
> import Debug.Trace
> import System.Console.GetOpt
> import System.Directory
> import System.Environment
> import System.Exit
> import System.IO
> import System.IO.Unsafe
> import System.Posix.Files
> import System.Process
> import Text.ParserCombinators.Parsec

> import Configuration
> import Version

TODO: shouldn't need to use IORef configurations... maybe try out
implicit parameters for this?

Configuration options:

> selectedOutputFormat = unsafePerformIO (newIORef printHumanReadableOutput)
> selectedMode = unsafePerformIO (newIORef traditionalMode)


> main = do 

>   (opts, dirs) <- parseCommandLine <$> getArgs

>   mapM (\o -> case o of
>           ForRM -> writeIORef selectedOutputFormat printDeleteFriendlyOutput
>           ZeroTerm -> writeIORef selectedOutputFormat printZeroTerminatedOutput
>           Help -> do { putStr (usageInfo ("lsdupes " ++ version ++ "\nCopyright 2006-2010 Ben Clifford. BSD-like licence") commandLineOptions); exitWith ExitSuccess }
>           INode -> writeIORef selectedMode inodeMode
>        ) opts

List files (map args to lists of files and fold, such that we keep
            the directory order specified in the args)

>   files <- listFiles (if dirs == [] then ["."] else dirs)

>   showProgress $ "Number of files: "++(show $ length files)

>   strimmedBySize <- dropSingletons <$> (partitionByHashM hashSize files)

>   showProgress $ "Number of file groups where more than one file has same size: "++(show $ length strimmedBySize)
>   showDebug $ "The largest comparison group is "++(show (foldr1 max (map length strimmedBySize)))


Now we have a bunch of comparison groups. Files in different comparison
groups are definitely different, so we only need to compare files in each
comparison group with entries in that same group.

>   mode <- readIORef selectedMode
>   mode strimmedBySize


> traditionalMode strimmedBySize = do

Now annotate every file in every comparison group with its md5sum, and then
generate a new partition based on that. In the old code, we end up with an
extra layer of nesting, but the repartitioning function can get rid of that
for us so that partitions look the same, type-wise, no matter how far down
the chain we are.

>   strimmedByMD5 <- dropSingletons <$> (refinePartitionByHashM hashMD5 strimmedBySize)

>   showProgress $ "Number of file groups where more than one file has same md5: "++(show $ length strimmedByMD5)

>   outputter <- readIORef selectedOutputFormat
>   outputter strimmedByMD5


> inodeMode strimmedBySize = do

>   partitionedByInode <- refinePartitionByHashDroppingSingletonsM hashInode strimmedBySize

>   showProgress $ "Number of inodes to probe: "++(show $ length partitionedByInode)

   print partitionedByInode -- remove before commit

partitionedByInode is then a partitioning of the original file space such
that all elements in a partition have the same inode (its the list of
filenames under the equivalent relation 'is same inode').

Now for each inode, take the md5sum of that inode. We need to give md5sum
a filename, but we can equally well pick any of the filenames in the
equivalence class as they all go to the same inode.

Somewhere around here, need to avoid md5summing stuff that doesn't need
md5summing - when any particular size has only one inode group, then we
can drop that size entirely (i.e. the strimmed-by-size, but against
inodes rather than against filenames)

But partitionedByInode at the moment has no size equivalence information
any more - refinePartitionByHashM loses that.

>   let annotateInodeWithMD5 equivalenceClass = (\md5 -> (equivalenceClass, md5)) <$> hashMD5 (head equivalenceClass)

>   inodesAnnotatedByMD5 <- mapM annotateInodeWithMD5 partitionedByInode

Now group all of the inodes by their MD5. We'll end up with partition under
the 'has same md5' relation.

>   let inodesGroupedByMD5 = sortAndGroupByAnnotation inodesAnnotatedByMD5

>   showProgress $ "Number of md5 groups: "++(show $ length inodesGroupedByMD5)

Then, for each equivalence class of MD5s:
for all except the first equivalence class of inodes, delete all the
files and replace them with hard links to the first inode.

>   let replaceFile toKeep toReplace = do {
>     putStrLn ("rm \""++toReplace++"\"") ;
>     putStrLn ("ln \""++toKeep++"\" \""++toReplace++"\"")
>   }

>   let processMD5InodeGroup g = do {
>     let {keep = head g;
>          keepFilename = head keep ; -- arbitrary filename from first group
>          tolink = tail g } ;
>     if length tolink > 0 then do
>         showDebug ("must replace "++(show tolink)++" with link to "++keepFilename )

This is not the right place to do output, I think, but its fine for
experimenting...

>         mapM_ (\l -> mapM_ (replaceFile keepFilename) l) tolink
>      else return ()
>   }

>   mapM_ processMD5InodeGroup inodesGroupedByMD5


===== The hash functions =====

The traditional hash functions have this meaning:
  If two files have different hash values, then the content is definitely
  different.
  If two files have the same hash values, then no assertion is made about
  their content.

The inode hash function has a different sense:
  If two files have the same hash value, then the content is definitely
  the same.
  If two files have different hash values, then no assertion is made about
  their content.

This second style of hash function has a different use. When attempting to
make all directory entries that contain the same content point at the
same inode (so that files do not disappear from the filesystem, but their
wasted space is freed), it is desirable to know which files point to different
inodes.


> fileToHashpair h f = (\s -> (f,s)) <$> (h f)


A hash based on content size. This is cheap because it requires access
only to the inode for a file, and not to any file content. For some
data collections it is very accurate.

> hashSize f = fileSize <$> getFileStatus f


A hash based on md5sum. This is more expensive to compute than hashSize.

> hashMD5 filename = do

this was for debugging:

>   showDebug $ "Taking md5 of "++filename

>   (inp, out, err, pid) <- runInteractiveProcess
>     md5sumCommand [filename] Nothing Nothing
>   output <- hGetLine out
>   let a = take 32 output
>   waitForProcess pid
>   hClose out
>   hClose inp
>   hClose err
>   return a


A hash that returns the deviceid and inode for a file.

> hashInode filename =
>    (\status -> (deviceID status, fileID status)) <$> getFileStatus filename


===== Partition operations =====

Drop any element of a list of lists which has only zero or a single member

> dropSingletons partition = filter (\ec -> length ec >= 2) partition


> sortAndGroupByAnnotation :: (Ord annotation, Eq annotation) => [(content,annotation)] -> [[content]]
> sortAndGroupByAnnotation l = let
>     s = sortBy (\(_,l) -> \(_,r) -> compare l r) l
>     g = groupBy (\(_,l) -> \(_,r) -> l==r) s

Chop off the annotation as we won't need it any more.

>     r = map (\l -> map (\(content,_) -> content) l) g

>   in r


Partition a list into a list of lists, grouping by the results of the
supplied hash function. This is monadic as the hash functions in lsdupes
perform IO operations.

> partitionByHashM hash list = do
>   annotatedFiles <- mapM (fileToHashpair hash) list
>   return $ sortAndGroupByAnnotation annotatedFiles

> refinePartitionByHashM hash list = do
>   annotatedFiles <- mapM (\l->mapM (fileToHashpair hash) l) list
>   let sortedAndGrouped = map (\x -> (sortAndGroupByAnnotation x)) annotatedFiles
>   return $ foldr (++) [] sortedAndGrouped

> refinePartitionByHashDroppingSingletonsM hash list = do

list is a partition: [[file]]

>   annotatedFiles <- mapM (\l->mapM (fileToHashpair hash) l) list

annotatedFiles is [[(annotation,file)]]

>   let sortedAndGrouped = map (\x -> (sortAndGroupByAnnotation x)) annotatedFiles

sortedAndGrouped is [[[file]]]

The outer list is according to the old-partition; and each of the inner ones
are pieces of the new partition.

For example, if we've partitioned by size previously and are now
partitioning by inode number, then each of the first-level list members will
contain everything about a particular file size. So then if we only end
up with a single inode-number in a particular file size, we know that
there are no duplicates for that file size, and can ignore it from now
on by dropping it here.

>   let sortedAndGroupedAndUniquified = dropSingletons sortedAndGrouped

   putStrLn $ "sortedAndGrouped: "++(show sortedAndGrouped)
   putStrLn $ "sortedAndGroupedAndUniquified: "++(show sortedAndGroupedAndUniquified)

>   return $ foldr (++) [] sortedAndGroupedAndUniquified

===== Commandline option handling =====

> data Flag = ForRM | ZeroTerm | Help | INode

> parseCommandLine args =
>   case parseOptions args of
>     (o,d,[]) -> (o,d)
>     (_,_,errs) -> error ("Errors in command line: "++(show errs))


> commandLineOptions = [ 
>              Option "h?" ["help"] (NoArg Help) "show help",
>              Option "r" ["for-rm"] (NoArg ForRM) "rm friendly output",
>              Option "0" ["zero"] (NoArg ZeroTerm) "NUL-terminated output",
>              Option "i" ["inode"] (NoArg INode) "inode mode"
>            ]

> parseOptions args = getOpt Permute commandLineOptions args 


===== Filesystem enumeration =====

Be very careful in this code to preserve original order.

> listFiles l = do

>   a <- listFilesInDir (head l)
>   b <- if (tail l) /= [] then listFiles (tail l) else return []
>   return (a ++ b)

> listFilesInDir d = do
>   l1 <- getDirectoryContents d
>   let l = filter (\s -> s /= "." && s /= ".." )  l1
>   fl <- filterM (\f -> do { s <- getSymbolicLinkStatus (d++"/"++f) ; return $ (( isRegularFile s ) && not (isSymbolicLink s)) }) l 
>   dl <- filterM (\f -> do { s <- getSymbolicLinkStatus (d++"/"++f) ; return $ isDirectory s }) l 
>   subs <- mapM (\sd -> listFilesInDir (d++"/"++sd)  )  dl
>   let r = (foldr (++) [] subs) ++ (map (\f -> d++"/"++f)) fl
>   return r


===== Output handling =====

> type NestedOutput = [[String]]

First helper functions for the different styles of output format:

> printUsingWriterMonad :: (NestedOutput -> Writer String a) -> NestedOutput -> IO ()
> printUsingWriterMonad pr o = putStr (execWriter (pr o))


Now the actual output formats:

> printHumanReadableOutput = printUsingWriterMonad $ renderHumanReadableOutput

> renderHumanReadableOutput o = do
>     tell "=== Human readable (?!) output ===\n"
>     mapM (\r -> do { tell "> "; tell $ show $ head r ; tell " duplicate of : "; tell $ show $ tail r ; tell "\n" }) o


In the delete friendly output, we always ignore the head of the lists,
because thats the one we want to keep. Use print rather than putStr on
the filename so that we get "double quotes" to preserve spaces.

> printDeleteFriendlyOutput = printUsingWriterMonad $ renderDeleteFriendlyOutput

> renderDeleteFriendlyOutput o = do
>     mapM (\r -> mapM (\f-> do { tell "rm -v " ; tell $ show f ; tell "\n"}) (tail r) ) o


> printZeroTerminatedOutput = printUsingWriterMonad $ renderZeroTerminatedOutput

> renderZeroTerminatedOutput o = do
>     mapM (\r -> mapM (\f-> do { tell f; tell ['\0'] }) (tail r) ) o


===== Progress and debugging =====

These should be controlled by verbosity flags. For now, debug is hard
coded off, and progress is hard-coded on.

> showDebug m = return () -- hPutStrLn stderr
> showProgress m = hPutStrLn stderr m


